from pathlib import Path
from allosaurus.lm.inventory import Inventory
from allosaurus.model import get_model_path
import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser('List language phone inventory')
    parser.add_argument('-l', '--lang', type=str,  default='ipa',         help='specify which language inventory to use for recognition. default "ipa" is to use all phone inventory')
    parser.add_argument('-m', '--model', type=str, default='latest',     help='specify which model inventory')
    parser.add_argument('-a', '--approximate', type=bool, default=False, help='the phone inventory can still hardly to cover all phones. You can use turn on this flag to map missing phones to other similar phones to recognize. The similarity is measured with phonological features')

    args = parser.parse_args()

    model_path = get_model_path(args.model)

    inventory = Inventory(model_path)

    if args.lang == 'ipa':
        print(' '.join(list(inventory.unit.id_to_unit.values())[1:]))
    else:
        lang = args.lang
        assert lang.lower() in inventory.lang_ids or lang.lower() in inventory.glotto_ids, f'language {args.lang} is not supported. Please verify it is in the language list'

        mask = inventory.get_mask(args.lang.lower(), approximation=args.approximate)

        unit = mask.target_unit
        print(' '.join(list(unit.id_to_unit.values()))[1:])

        if args.approximate:
            mask.print_maps()


from pathlib import Path
from allosaurus.lm.inventory import Inventory
from allosaurus.lm.unit import write_unit
from allosaurus.model import get_model_path
import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser('Write out current phone file')
    parser.add_argument('-l', '--lang',   type=str, required=True, help='specify which language inventory to update.')
    parser.add_argument('-m', '--model',  type=str, default='latest', help='specify which model inventory')
    parser.add_argument('-o', '--output', type=str, required=True, help='write out your current phone file.')
    parser.add_argument('-f', '--format', type=str, default='simple', choices=['simple', 'kaldi'], help='select your output format')

    args = parser.parse_args()

    model_path = get_model_path(args.model)

    inventory = Inventory(model_path)

    lang = args.lang

    unit = inventory.get_unit(lang)
    write_unit(unit, args.output, args.format)


from allosaurus.am.utils import *
from pathlib import Path
from allosaurus.audio import read_audio
from allosaurus.pm.factory import read_pm
from allosaurus.am.factory import read_am
from allosaurus.lm.factory import read_lm
from allosaurus.bin.download_model import download_model
from allosaurus.model import resolve_model_name, get_all_models
from argparse import Namespace

def read_recognizer(inference_config_or_name='latest'):

    # download specified model automatically if no model exists
    if len(get_all_models()) == 0:
        download_model('latest')

    # create default config if input is the model's name
    if isinstance(inference_config_or_name, str):
        model_name = resolve_model_name(inference_config_or_name)
        inference_config = Namespace(model=model_name, device_id=-1, lang='ipa', approximate=False, prior=None)
    else:
        assert isinstance(inference_config_or_name, Namespace)
        inference_config = inference_config_or_name

    model_path = Path(__file__).parent / 'pretrained' / inference_config.model

    if inference_config.model == 'latest' and not model_path.exists():
        download_model(inference_config)

    assert model_path.exists(), f"{inference_config.model} is not a valid model"

    # create pm (pm stands for preprocess model: audio -> feature etc..)
    pm = read_pm(model_path, inference_config)

    # create am (acoustic model: feature -> logits )
    am = read_am(model_path, inference_config)

    # create lm (language model: logits -> phone)
    lm = read_lm(model_path, inference_config)

    return Recognizer(pm, am, lm, inference_config)

class Recognizer:

    def __init__(self, pm, am, lm, config):

        self.pm = pm
        self.am = am
        self.lm = lm
        self.config = config

    def recognize(self, filename, lang_id='ipa', topk=1, emit=1.0, timestamp=False):
        # recognize a single file

        assert str(filename).endswith('.wav'), "only wave file is supported in allosaurus"

        # load wav audio
        audio = read_audio(filename)

        # extract feature
        feat = self.pm.compute(audio)

        # add batch dim
        feats = np.expand_dims(feat, 0)
        feat_len = np.array([feat.shape[0]], dtype=np.int32)

        tensor_batch_feat, tensor_batch_feat_len = move_to_tensor([feats, feat_len], self.config.device_id)

        tensor_batch_lprobs = self.am(tensor_batch_feat, tensor_batch_feat_len)

        if self.config.device_id >= 0:
            batch_lprobs = tensor_batch_lprobs.cpu().detach().numpy()
        else:
            batch_lprobs = tensor_batch_lprobs.detach().numpy()

        token = self.lm.compute(batch_lprobs[0], lang_id, topk, emit=emit, timestamp=timestamp)
        return token


from allosaurus.lm.inventory import *
from pathlib import Path
from itertools import groupby
import numpy as np

class PhoneDecoder:

    def __init__(self, model_path, inference_config):
        """
        This class is an util for decode both phones and words

        :param model_path:
        """

        # lm model path
        self.model_path = Path(model_path)

        self.config = inference_config

        # create inventory
        self.inventory = Inventory(model_path, inference_config)

        self.unit = self.inventory.unit

    def compute(self, logits, lang_id=None, topk=1, emit=1.0, timestamp=False):
        """
        decode phones from logits

        :param logits: numpy array of logits
        :param emit: blank factor
        :return:
        """

        # apply mask if lang_id specified, this is to restrict the output phones to the desired phone subset

        mask = self.inventory.get_mask(lang_id, approximation=self.config.approximate)

        logits = mask.mask_logits(logits)

        emit_frame_idx = []

        cur_max_arg = -1

        # find all emitting frames
        for i in range(len(logits)):

            logit = logits[i]
            logit[0] /= emit

            arg_max = np.argmax(logit)

            # this is an emitting frame
            if arg_max != cur_max_arg and arg_max != 0:
                emit_frame_idx.append(i)
                cur_max_arg = arg_max

        # decode all emitting frames
        decoded_seq = []
        for idx in emit_frame_idx:
            logit = logits[idx]
            exp_prob = np.exp(logit - np.max(logit))
            probs = exp_prob / exp_prob.sum()

            top_phones = logit.argsort()[-topk:][::-1]
            top_probs = sorted(probs)[-topk:][::-1]

            stamp = f"{self.config.window_shift*idx:.3f} {self.config.window_size:.3f} "

            if topk == 1:

                phones_str = ' '.join(mask.get_units(top_phones))
                if timestamp:
                    phones_str = stamp + phones_str

                decoded_seq.append(phones_str)
            else:
                phone_prob_lst = [f"{phone} ({prob:.3f})" for phone, prob in zip(mask.get_units(top_phones), top_probs)]
                phones_str = ' '.join(phone_prob_lst)

                if timestamp:
                    phones_str = stamp + phones_str

                decoded_seq.append(phones_str)

        if timestamp:
            phones = '\n'.join(decoded_seq)
        elif topk == 1:
            phones = ' '.join(decoded_seq)
        else:
            phones = ' | '.join(decoded_seq)

        return phones


from allosaurus.am.utils import move_to_tensor, torch_save
from allosaurus.am.criterion import read_criterion
from allosaurus.am.optimizer import read_optimizer
from allosaurus.am.reporter import Reporter
import editdistance
import numpy as np
import torch
from itertools import groupby
from allosaurus.model import get_model_path
import os
import json

class Trainer:

    def __init__(self, model, train_config):

        self.model = model
        self.train_config = train_config

        self.device_id = self.train_config.device_id

        # criterion, only ctc currently
        self.criterion = read_criterion(train_config)

        # optimizer, only sgd currently
        self.optimizer = read_optimizer(self.model, train_config)

        # reporter to write logs
        self.reporter = Reporter(train_config)

        # best per
        self.best_per = 100.0

        # intialize the model
        self.model_path = get_model_path(train_config.new_model)

        # counter for early stopping
        self.num_no_improvement = 0


    def sum_edit_distance(self, output_ndarray, output_lengths_ndarray, token_ndarray, token_lengths_ndarray):
        """
        compute SUM of ter in this batch

        """

        error_cnt_sum = 0.0

        for i in range(len(token_lengths_ndarray)):
            target_list = token_ndarray[i, :token_lengths_ndarray[i]].tolist()
            logit = output_ndarray[i][:output_lengths_ndarray[i]]

            raw_token = [x[0] for x in groupby(np.argmax(logit, axis=1))]
            decoded_token = list(filter(lambda a: a != 0, raw_token))

            error_cnt_sum += editdistance.distance(target_list, decoded_token)

        return error_cnt_sum


    def step(self, feat_batch, token_batch):

        # prepare torch tensors from numpy arrays
        feat_tensor, feat_lengths_tensor = move_to_tensor(feat_batch, self.device_id)
        token_tensor, token_lengths_tensor = move_to_tensor(token_batch, self.device_id)

        #print(feat_tensor)
        #print(feat_lengths_tensor)
        output_tensor = self.model(feat_tensor, feat_lengths_tensor)

        #print(output_tensor)
        #print(token_tensor)
        #print(token_lengths_tensor)

        loss = self.criterion(output_tensor, feat_lengths_tensor, token_tensor, token_lengths_tensor)
        #print(loss.item())

        # extract numpy format for edit distance computing
        output_ndarray = output_tensor.cpu().detach().numpy()
        feat_ndarray, feat_lengths_ndarray = feat_batch
        token_ndarray, token_lengths_ndarray = token_batch

        phone_error_sum = self.sum_edit_distance(output_ndarray, feat_lengths_ndarray, token_ndarray,
                                                 token_lengths_ndarray)

        phone_count = sum(token_lengths_ndarray)

        return loss, phone_error_sum, phone_count


    def train(self, train_loader, validate_loader):

        self.best_per = 100.0

        batch_count = len(train_loader)

        for epoch in range(self.train_config.epoch):

            # shuffle
            train_loader.shuffle()

            # set to the training mode
            self.model.train()

            # reset all stats
            all_phone_count = 0.0
            all_loss_sum = 0.0
            all_phone_error_sum = 0.0

            # training loop
            for ii in range(batch_count):

                self.optimizer.zero_grad()

                feat_batch, token_batch = train_loader.read_batch(ii)

                # forward step
                loss_tensor, phone_error_sum, phone_count = self.step(feat_batch, token_batch)

                # backprop and optimize
                loss_tensor.backward()

                torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.train_config.grad_clip)

                self.optimizer.step()

                # update stats
                loss_sum = loss_tensor.item()
                all_phone_count += phone_count
                all_loss_sum += loss_sum
                all_phone_error_sum += phone_error_sum

                if ii % self.train_config.report_per_batch == 0:
                    message = f'epoch[batch]: {epoch:02d}[{ii:04d}] | train loss {all_loss_sum/all_phone_count:0.5f} train per {all_phone_error_sum / all_phone_count:0.5f}'
                    self.reporter.write(message)

                    # reset all stats
                    all_phone_count = 0.0
                    all_loss_sum = 0.0
                    all_phone_error_sum = 0.0


            # evaluate this model
            validate_phone_error_rate = self.validate(validate_loader)

            self.reporter.write(f"epoch{epoch} | validate per : {validate_phone_error_rate:0.5f}")
            if validate_phone_error_rate <= self.best_per:
                self.best_per = validate_phone_error_rate
                self.num_no_improvement = 0
                self.reporter.write("saving model")

                model_name = f"model_{validate_phone_error_rate:0.5f}.pt"

                # save model
                torch_save(self.model, self.model_path / model_name)

                # overwrite the best model
                torch_save(self.model, self.model_path / 'model.pt')

            else:
                self.num_no_improvement += 1

                if self.num_no_improvement >= 3:
                    self.reporter.write("no improvements for several epochs, early stopping now")
                    break

        # close reporter stream
        self.reporter.close()


    def validate(self, validate_loader):

        self.model.eval()

        batch_count = len(validate_loader)

        all_phone_error_sum = 0
        all_phone_count = 0

        # validation loop
        for ii in range(batch_count):

            self.optimizer.zero_grad()

            feat_batch, token_batch = validate_loader.read_batch(ii)

            # one step
            loss_tensor, phone_error_sum, phone_count = self.step(feat_batch, token_batch)

            # update stats
            all_phone_error_sum += phone_error_sum
            all_phone_count += phone_count

        return all_phone_error_sum/all_phone_count



from allosaurus.model import delete_model
import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser('an allosaurus util to delete model')
    parser.add_argument('-m', '--model', required=True, help='model name to be deleted')

    args = parser.parse_args()
    delete_model(args.model)


from pathlib import Path
import shutil

def get_all_models():
    """
    get all local models

    :return:
    """
    model_dir = Path(__file__).parent / 'pretrained'
    models = list(sorted(model_dir.glob('*'), reverse=True))

    #assert len(models) > 0, "No models are available, you can maually download a model with download command or just run inference to download the latest one automatically"

    return models

def get_model_path(model_name):
    """
    get model path by name, verify its a valid path

    :param model_name: str
    :return: model path
    """

    model_dir = Path(__file__).parent / 'pretrained'

    resolved_model_name = resolve_model_name(model_name)

    assert resolved_model_name != "none", model_name+" is not a valid model name. please check by list_model"

    return model_dir / resolved_model_name

def copy_model(src_model_name, tgt_model_name):
    """
    copy a model to a new model

    :param src_model_name:
    :param tgt_model_name:
    :return:
    """

    # verify the source path is not empty
    src_model_path = get_model_path(src_model_name)

    # verify the target path is empty
    model_dir = Path(__file__).parent / 'pretrained'
    tgt_model_path = model_dir / tgt_model_name

    assert not tgt_model_path.exists(), \
        "provided model name "+tgt_model_name+" has already exist. Consider another name or delete the existing one"

    shutil.copytree(str(src_model_path), str(tgt_model_path))

def delete_model(model_name):

    model_path = get_model_path(model_name)

    answer = input(f"you will delete {model_path}? [Y|N]")
    if answer.lower() in ['y', 'yes', 'true']:
        print("deleting ", model_path)
        shutil.rmtree(str(model_path))


def resolve_model_name(model_name='latest'):
    """
    select the model

    :param model_name:
    :return:
    """

    models = get_all_models()

    # get the latest model in local
    if model_name == 'latest':
        return models[0].name

    # match model name
    for model in models:
        if model.name == model_name:
            return model_name

    return "none"


import argparse
from pathlib import Path
from allosaurus.model import resolve_model_name
from allosaurus.audio import read_audio
from allosaurus.pm.factory import read_pm
from allosaurus.pm.kdict import KaldiWriter
from tqdm import tqdm

def prepare_feature(data_path, model):

    model_path = Path(__file__).parent.parent / 'pretrained' / model

    # create pm (pm stands for preprocess model: audio -> feature etc..)
    pm = read_pm(model_path, None)

    # data path should be pointing the absolute path
    data_path = data_path.absolute()

    # writer for feats
    feat_writer = KaldiWriter(data_path / 'feat')

    # writer for the shape of each utterance
    # format: utt_id shape[0] shape[1]
    shape_writer = open(data_path / 'shape', 'w')

    for line in tqdm(open(data_path / 'wave', 'r', encoding='utf-8').readlines()):
        fields = line.strip().split()
        utt_id = fields[0]
        audio_path = fields[1]

        assert Path(audio_path).exists(), audio_path+" does not exist!"

        audio = read_audio(audio_path)

        # extract feature
        feat = pm.compute(audio)

        # write shape
        shape_writer.write(f'{utt_id} {feat.shape[0]} {feat.shape[1]}\n')

        feat_writer.write(utt_id, feat)

    feat_writer.close()
    shape_writer.close()

if __name__ == '__main__':

    parser = argparse.ArgumentParser('allosaurus tool to extract audio feature for fine-tuning')
    parser.add_argument('--path', required=True, type=str, help='path to the directory containing the wave file')
    parser.add_argument('--model', type=str, default='latest', help='specify the model you want to fine-tune')

    args = parser.parse_args()
    data_path = Path(args.path)

    wave_path = data_path / 'wave'

    assert wave_path.exists(), "the path directory should contain a wave file, please check README.md for details"

    # resolve model's name
    model_name = resolve_model_name(args.model)
    if model_name == "none":
        print("Model ", model_name, " does not exist. Please download this model or use an existing model in list_model")
        exit(0)

    args.model = model_name

    # extract feature
    prepare_feature(data_path, args.model)






import json
from allosaurus.lm.mask import *


class Inventory:

    def __init__(self, model_path, inference_config=None):

        self.model_path = model_path

        self.lang_names = []
        self.lang_ids = []
        self.glotto_ids = []

        self.lang2phonefile = dict()

        self.inference_config = inference_config

        # load all available inventories
        langs = json.load(open(self.model_path / 'inventory' / 'index.json', 'r', encoding='utf-8'))

        # load all phones list
        self.unit = read_unit(str(self.model_path / 'phone.txt'))

        for lang in langs:

            lang_name = lang['LanguageName']
            iso_id = lang['ISO6393']
            glotto_id = lang['GlottoCode']
            phone_text = lang['phonelists']

            self.lang_names.append(lang_name.lower())
            self.lang_ids.append(iso_id.lower())
            self.glotto_ids.append(glotto_id)

            # register both iso_id and glottocode id
            self.lang2phonefile[iso_id.lower()] = phone_text
            self.lang2phonefile[glotto_id.lower()] = phone_text

    def get_unit(self, lang_id):
        """
        load a unit specified by the lang_id

        Args:
            lang_id: ISO id

        Returns:
        """

        # we can also specify a unit path as lang_id, makes it easier to customize
        if lang_id not in self.lang2phonefile:

            if not Path(lang_id).exists():
                assert lang_id in self.lang2phonefile, "Language "+lang_id+" is not available !"
            else:
                target_unit = read_unit(str(lang_id))
                return target_unit


        # search customized file first, if not exist use the default one.
        updated_unit_file = self.model_path / 'inventory' / ('updated_'+self.lang2phonefile[lang_id])
        if updated_unit_file.exists():
            unit_file = updated_unit_file
        else:
            unit_file = self.model_path / 'inventory' / self.lang2phonefile[lang_id]

        target_unit = read_unit(str(unit_file))

        return target_unit

    def update_unit(self, lang_id, unit_file):
        """
        update the existing unit with a new unit file

        Args:
            lang_id:
            unit_file:

        Returns:

        """

        assert lang_id in self.lang2phonefile, "Language "+lang_id+" is not available !"


        # load the new unit file and validate its format
        new_unit = read_unit(unit_file)

        # the model path it should be stored
        updated_unit_file = self.model_path / 'inventory' / ('updated_'+self.lang2phonefile[lang_id])

        # save the new file
        write_unit(new_unit, updated_unit_file)

    def restore_unit(self, lang_id):
        """
        restore the original phone units

        Args:
            lang_id:

        Returns:

        """

        assert lang_id in self.lang2phonefile, "Language "+lang_id+" is not available !"

        # the updated unit file
        updated_unit_file = self.model_path / 'inventory' / ('updated_'+self.lang2phonefile[lang_id])

        # check whether it has an updated file
        assert updated_unit_file.exists(), "language "+lang_id+" does not have any customized inventory."

        # delete this file
        updated_unit_file.unlink()

    def get_mask(self, lang_id=None, approximation=False):

        # use its unit as the mask if the recognition target is the entire inventory
        if lang_id is None or lang_id == 'ipa':
            target_unit = self.unit
        else:
            target_unit = self.get_unit(lang_id)

        return UnitMask(self.unit, target_unit, approximation, self.inference_config)


import unittest
from pathlib import Path
import requests

class TestModel(unittest.TestCase):

    def test_latest_available(self):
        req = requests.head('https://www.pyspeech.com/static/model/recognition/allosaurus/latest.tar.gz')
        self.assertTrue(req.status_code==200)

if __name__ == '__main__':
    unittest.main()


from pathlib import Path
from allosaurus.model import get_model_path
from allosaurus.lm.inventory import Inventory
import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser('List language phone inventory')
    parser.add_argument('-l', '--lang',  type=str, default='ipa',         help='specify which language inventory to use for recognition. default "ipa" is to use all phone inventory')
    parser.add_argument('-m', '--model', type=str, default='latest',     help='specify which model inventory')

    args = parser.parse_args()
    model_path = get_model_path(args.model)

    inventory = Inventory(model_path)

    print("Available Languages")
    for lang_id, glotto_id, lang_name in zip(inventory.lang_ids, inventory.glotto_ids, inventory.lang_names):
        lang_name = lang_name.encode('ascii', 'ignore')
        print('- ISO639-3: ', lang_id, 'Glotto Code', glotto_id, ' name: ', lang_name)



from pathlib import Path
from allosaurus.lm.inventory import Inventory
from allosaurus.model import get_model_path
import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser('Update language inventory')
    parser.add_argument('-l', '--lang',  type=str, required=True, help='specify which language inventory to update.')
    parser.add_argument('-m', '--model', type=str, default='latest', help='specify which model inventory')
    parser.add_argument('-i', '--input', type=str, required=True, help='your new inventory file')

    args = parser.parse_args()

    model_path = get_model_path(args.model)

    inventory = Inventory(model_path)

    lang = args.lang

    # verify lang is not ipa as it is an alias to the entire inventory
    assert args.lang != 'ipa', "ipa is not a proper lang to update. use list_lang to find a proper language"

    assert lang.lower() in inventory.lang_ids or lang.lower() in inventory.glotto_ids, f'language {args.lang} is not supported. Please verify it is in the language list'

    new_unit_file = Path(args.input)

    # check existence of the file
    assert new_unit_file.exists(), args.input+' does not exist'

    # update this new unit
    inventory.update_unit(lang, new_unit_file)


from allosaurus.am.dataset import AllosaurusDataset
import numpy as np

def read_loader(data_path, train_config):
    """
    create a dataloader for data_path

    :param data_path:
    :param train_config:
    :return:
    """

    return AllosaurusLoader(data_path, train_config)


class AllosaurusLoader:

    def __init__(self, data_path, train_config):

        self.train_config = train_config

        self.dataset = AllosaurusDataset(data_path)

        self.batch_lst = []

        self._prepare_batch()

    def __len__(self):
        return len(self.batch_lst)

    def close(self):
        self.dataset.close()

    def shuffle(self):
        np.random.shuffle(self.batch_lst)

    def read_batch(self, batch_idx):
        assert batch_idx < len(self.batch_lst), "batch_idx "+str(batch_idx)+" is too large!!"

        batch = self.batch_lst[batch_idx]

        return self._collate_batch(batch)

    def _collate_batch(self, batch):

        feat_lst = []
        token_lst = []

        for idx in batch:
            feat, token = self.dataset[idx]
            feat_lst.append(feat)
            token_lst.append(token)

        feat, feat_lengths = self._collate_feat(feat_lst)
        token, token_lengths = self._collate_token(token_lst)

        return (feat, feat_lengths), (token, token_lengths)

    def _collate_feat(self, feat_lst):

        batch_size = len(feat_lst)
        frame_size = feat_lst[0].shape[0]
        feat_size = feat_lst[0].shape[1]

        # collate feats
        feat_lengths = np.zeros(batch_size, dtype=np.int32)
        feat_tensor = np.zeros([batch_size, frame_size, feat_size], dtype=np.float32)
        for i, feat in enumerate(feat_lst):

            feat_tensor[i,:len(feat)] = feat
            feat_lengths[i] = len(feat)

        return feat_tensor, feat_lengths


    def _collate_token(self, token_lst):

        batch_size = len(token_lst)
        token_size = max([len(token) for token in token_lst])

        token_tensor = np.zeros([batch_size, token_size], dtype=np.int32)
        token_lengths = np.zeros(batch_size, dtype=np.int32)

        for i, token in enumerate(token_lst):
            token_tensor[i, :len(token)] = token
            token_lengths[i] = len(token)

        return token_tensor, token_lengths


    def _prepare_batch(self):

        batch = []
        batch_frame_size = 0

        for i in range(len(self.dataset)):
            utt_id = self.dataset.utt_ids[i]

            frame_size = self.dataset.utt2shape[utt_id][0]

            # batch frame is large enough
            if batch_frame_size + frame_size >= self.train_config.batch_frame_size:

                # commit current batch to the list
                self.batch_lst.append(batch)

                # reset frame size
                batch_frame_size = 0

                # reset batch
                batch = []

            batch_frame_size += frame_size
            batch.append(i)

        # commit the last batch if it is a valid batch
        if len(batch) > 0 and batch_frame_size < self.train_config.batch_frame_size:
            self.batch_lst.append(batch)


from allosaurus.pm.kdict import read_matrix
from pathlib import Path
from torch.utils.data import Dataset
import numpy as np

class AllosaurusDataset(Dataset):

    def __init__(self, data_path):
        self.data_path = Path(data_path)

        required_files = ['feat.scp', 'token', 'feat.ark', 'shape']

        for required_file in required_files:
            assert (self.data_path / required_file).exists(), required_file+" does not exist, please run the preparation before fine-tuning"

        # read all tokens
        self.utt2token = {}
        self._read_token()

        # read all features and their shapes
        self.utt2offset = {}
        self.utt2shape = {}
        self.ark = None
        self._read_feat()

        # extract all valid utt_ids
        token_utt_set = set(self.utt2token.keys())
        feat_utt_set = set(self.utt2offset.keys())
        shape_utt_set = set(self.utt2shape.keys())
        self.utt_ids = list(set.intersection(token_utt_set, feat_utt_set, shape_utt_set))

        # sort all ids based on their their shape
        self.utt_ids.sort(key=lambda utt_id: self.utt2shape[utt_id][0], reverse=True)

    def __len__(self):
        return len(self.utt_ids)

    def __getitem__(self, item):

        utt_id = self.utt_ids[item]

        token = self.utt2token[utt_id]

        offset = self.utt2offset[utt_id]

        self.ark.seek(offset)

        feature = read_matrix(self.ark, np.float32)

        return (feature, token)

    def close(self):

        if self.ark:
            self.ark.close()

    def _read_token(self):
        """
        load token from file

        :param token_path:
        :return:
        """

        token_reader = open(str(self.data_path / 'token'), 'r', encoding='utf-8')

        self.utt2token = {}

        for line in token_reader:
            fields = line.strip().split()
            utt_id = fields[0]
            tokens = list(map(int, fields[1:]))

            # reject empty token or too long token
            if len(tokens) == 0 or len(tokens) > 1000:
                continue

            self.utt2token[utt_id] = tokens


    def _read_feat(self):
        """
        load offsets from feat.scp

        :return:
        """

        #####################################################################
        # read feature
        #####################################################################

        feat_reader = open(str(self.data_path / 'feat.scp'), 'r')
        self.utt2offset = {}
        for line in feat_reader:
            fields = line.strip().split()

            assert len(fields) == 2, " feat.scp should only contain two fields"
            utt_id = fields[0]

            feat = fields[1]
            p = feat.rfind(":")

            assert p >= 0, " offset pointer not found"
            offset = int(feat[p+1:])
            self.utt2offset[utt_id] = offset

        feat_reader.close()

        self.ark = open(self.data_path / 'feat.ark', 'rb')

        #####################################################################
        # read shape
        #####################################################################

        shape_reader = open(str(self.data_path / 'shape'), 'r')

        for line in shape_reader:
            fields = line.strip().split()
            utt_id = fields[0]

            shape = (int(fields[1]), int(fields[2]))

            self.utt2shape[utt_id] = shape

        shape_reader.close()